==> mdcached-1.0.2.tgz: mdcached-1.0.2/mdcached/mc_protocol.h <==
/*-
 * This file is a part of MDCacheD
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: mc_protocol.h 411 2012-01-20 15:47:47Z ivoras $
 */

/*
 * MDCacheD protocol-related definitions.
 */

#ifndef MC_PROTOCOL_H
#define MC_PROTOCOL_H

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE (!TRUE)
#endif

#define MC_PROTOCOL_VERSION	1
#define MC_PROTOCOL_VERSION_ID	"$Id: mc_protocol.h 411 2012-01-20 15:47:47Z ivoras $"

/*
 * Depends on: nothing.
 * This file is usable as a client-side header if the MDCACHED_SERVER symbol
 * is not defined.
 */


/*
 * This defines a network protocol so anything that goes over the wire must be
 * declared here.
 */

#define MC_DEFAULT_INET_PORT 	1222
#define MC_BASE_UNIX_SOCKET	"/tmp/mdcached"
#define MC_XSTR(s) #s
#define MC_GET_UNIX_SOCKET_PORT(port)	(MC_BASE_UNIX_SOCKET "." MC_XSTR(port))
#define MC_DEFAULT_UNIX_SOCKET MC_GET_UNIX_SOCKET_PORT(MC_DEFAULT_INET_PORT)

/**
 * Cutoff time in seconds. Time specifications less than this number
 * of seconds will be treated as relative to current time.
 */
#define REL_TIME_CUTOFF (60*60*24*30)

#ifdef TAGS_64BIT
#warning Record tags redefined as 64-bit!
typedef int64_t tag_value_t;		/* tag names and tag values are integers */
typedef int64_t tag_key_t;
#else
typedef int32_t tag_key_t;
typedef int32_t tag_value_t;
#endif

#define MCMD_DATA_ENTRY	MCMD_PUT_DATA

enum MCMD {
	/* commands */
	MCMD_INVALID			= 0,
	MCMD_HANDSHAKE			= 1,
	MCMD_PUT_DATA			= 2,
	MCMD_DEL_DATA_NAME		= 3,
	MCMD_DEL_DATA_BY_TAG_VALUES	= 4,
	MCMD_GET_DATA_NAME		= 5,
	MCMD_GET_DATA_BY_TAG_VALUES	= 6,
	MCMD_GET_ALL_KEYS		= 7,
	MCMD_ATOMIC_CMPSET		= 8,
	MCMD_ATOMIC_ADD			= 9,
	MCMD_ATOMIC_FETCHADD		= 10,
	MCMD_ATOMIC_READANDCLEAR	= 11,
	MCMD_MPUT			= 12,
	MCMD_MGET			= 13,
	MCMD_MDEL			= 14,
	MCMD_TSTACK_PUSH		= 15,
	MCMD_TSTACK_POP			= 16,

	/* responses */
	MCRESP_STATUS			= 100,
	MCRESP_HANDSHAKE		= 101,
	MCRESP_DATA			= 102,
	MCRESP_MULTIDATA		= 103,
	MCRESP_MULTISTRING		= 104,
	MCRESP_ATOMIC			= 105,
	MCRESP_TSTACK_PUSH		= 106,
	MCRESP_TSTACK_POP		= 107
};


enum MCSTATUS {
	MCSTATUS_OK,
	MCSTATUS_ERROR
};


enum MCREASON {
	MCREASON_OK = 0,
	MCREASON_INVALID,
	MCREASON_ENDIANESS,
	MCREASON_VERSION,
	MCREASON_INVALIDCMD,
	MCREASON_CACHE,
	MCREASON_NOTFOUND,
	MCREASON_SERVER,
	MCREASON_PARAM,
	MCREASON_ENOMEM
};

/* Header flags and definitions */
#define MC_ENDIAN_SIGNATURE 	0x1234abcd
#define MC_VER_MAJOR 		0
#define MC_VER_MINOR 		1
#define MC_FLAG_REPLY		0x01

/* Command flags */
#define MCMD_FLAG_ATOMICALL	0x1

/*
 * Command and response packets use the same over the wire format, starting with
 * mc_header and then continuing with the data.
 * Data layout in packets:
 *  - header (mc_header)
 *  - packet specific struct (mc_handshake | mc_add_data | ...)
 *  - variable-length data specific to command
 *    General layout of the variable section is: (not applicable fields are skipped)
 *        - tags (num_tags, tags)
 *        - name (name_length, name)
 *        - data (data_length, data)
 *  - name and data are variable-length, non-0-terminated binary strings.
 *  - name is limited to 64 kB, data to 4 GB.
 */

#ifndef __packed
 #define __packed        __attribute__((__packed__))
#endif


/* Record tag */
struct __packed mc_tag {
	tag_key_t key;
	tag_value_t value;
};


/**
 * Key name simple definition
 */
struct __packed mc_key_name {
	uint16_t name_size;
	char name[];
};


/**
 * Key data simple definition
 */
struct __packed mc_key_data {
	uint32_t data_size;
	char data[];
};


/**
 * Command packet header. This tructure is at the beginning of *every* command
 * packet and response in the client-server protocol.
 */
struct __packed mc_header {
	uint8_t mcmd;		/* Command type */
	uint8_t flags;		/* Command flags */
	uint16_t seq;		/* Command id (sequential) */
	uint32_t total_size;	/* Total size of the command packet, including this header */
};


/** "Handshake" command packet */
struct __packed mc_handshake {
	struct mc_header header;
	int32_t endian_signature;
	uint16_t proto_ver_major;
	uint16_t proto_ver_minor;
};


/** A "simple request" packet. Doesn't carry any data. */
struct __packed mc_simple {
	struct mc_header header;
	uint32_t flags;
	uint32_t param;
};


/**
 * "Add data" command packet. The payload (data field) contains an array of
 * struct mc_tag (n_tags items), followed by the name (of the length name_size),
 * followed by the record value/data (of the length data_size).
 */
struct __packed mc_data_entry {
	struct mc_header header;
	uint16_t n_tags;
	uint16_t name_size;
	uint32_t data_size;
	uint32_t exptime;
	char data[];			/* Dummy placeholder */
};

#define mc_add_data mc_data_entry
#define mc_resp_data mc_data_entry

static inline struct mc_tag* mc_data_entry_tags(struct mc_data_entry *md) {
	return (struct mc_tag*)(md->data);
}

static inline unsigned mc_data_entry_n_tags(struct mc_data_entry *md) {
	return md->n_tags;
}

static inline char* mc_data_entry_key(struct mc_data_entry *md) {
	return md->data + md->n_tags * sizeof(struct mc_tag);
}

static inline unsigned mc_data_entry_key_size(struct mc_data_entry *md) {
	return md->name_size;
}

static inline char* mc_data_entry_value(struct mc_data_entry *md) {
	return md->data + md->n_tags * sizeof(struct mc_tag) + md->name_size;
}

static inline unsigned mc_data_entry_value_size(struct mc_data_entry *md) {
	return md->data_size;
}

/** "Delete data by name" command packet */
struct __packed mc_del_data_name {
	struct mc_header header;
	uint16_t name_size;
	uint16_t _unused;
	char data[];
};


/** "Delete data by tag values" command packet */
struct __packed mc_del_data_by_tag_values {
	struct mc_header header;
	uint32_t n_values;
	tag_key_t tag_key;
	char data[];
};


/** "Get data by name" command packet */
struct __packed mc_get_data_name {
	struct mc_header header;
	uint16_t name_size;
	uint16_t _unused;
	char data[];
};


/** "Get data by tags" command packet */
struct __packed mc_get_data_by_tag_values {
	struct mc_header header;
	uint32_t n_values;
	tag_key_t tag_key;
	char data[];
};


/** A common command packet for simple atomic ops */
struct __packed mc_atomic_op {
	struct mc_header header;
	int64_t arg1;
	int64_t arg2;
	uint16_t name_size;
	uint16_t _unused;
	char name[];
};


/**
 * A multi-put operation: inserts multiple records into the datastore. The data
 * contains n_records repetitions of mc_data_entry records.
 */
struct __packed mc_mput {
	struct mc_header header;
	uint16_t n_records;
	uint16_t flags;
	char data[];
};


/**
 * A multi-get operation: gets multiple records by their names. The data
 * contains n_names repetitions of (uint16_t name length, name bytes)
 */
struct __packed mc_mget {
	struct mc_header header;
	uint16_t n_names;
	uint16_t flags;
	char data[];
};


/**
 * A multi-delete operation: deletes multiple records by their names. The data
 * contains n_names repetitions of (uint16_t name length, name bytes)
 */
struct __packed mc_mdel {
	struct mc_header header;
	uint16_t n_names;
	uint16_t flags;
	char data[];
};


/**
 * The tag-stack push operation. Data contains n_tags repetitions of the struct
 * mc_tag followed by the data of data_size bytes. Response is mc_resp_tstack_push.
 */
struct __packed mc_tstack_push {
	struct mc_header header;
	uint16_t flags;
	uint16_t n_tags;
	uint32_t data_size;
	char data[];
};


/**
 * The tag-stack pop operation. The tag (key and value) references the virtual
 * stack from which the record is popped (returned & removed). Response is mc_resp_
 */
struct __packed mc_tstack_pop {
	struct mc_header header;
	struct mc_tag tag;
	uint16_t flags;
};


/**
 * Status response packet
 */
struct __packed mc_resp_status {
	struct mc_header header;
	uint16_t status;
	uint16_t data;
};


/**
 * "Multi data" response - when a request matches multiple data items. The data
 * payload contains concatenated struct mc_data_entry blobs, unaligned.
 */
struct __packed mc_resp_multidata {
	struct mc_header header;
	uint32_t n_records;
	char data[];
};


/**
 * "Multi string" response - when a command returns an arbitrary list of strings.
 * The data payload consts of a concatenated list of (uint16_t name_len, char name[name_len])
 * records, unaligned. The length is declared as short unsigned integer as this
 * response type is intended to return record keys (key names).
 */
struct __packed mc_resp_multistring {
	struct mc_header header;
	uint32_t n_records;
	uint32_t data_size;
	char data[];
};


/**
 * A simple reply to an atomic op.
 */
struct __packed mc_resp_atomic {
	struct mc_header header;
	int64_t result;
};


/**
 * The response to struct mc_tstack_push. It contains the unique name generated
 * for the record.
 */
struct __packed mc_resp_tstack_push {
	struct mc_header header;
	uint16_t name_size;
	char name[];
};


/**
 * Update the common fields in the packet.
 */
static inline void
mc_init_packet(void *packet, const unsigned int size, const unsigned short int seq, const short int cmd)
{
	struct mc_header *pkt = (struct mc_header*) packet;
	pkt->mcmd = cmd;
	pkt->flags = 0;
	pkt->seq = seq;
	pkt->total_size = size;
}


/**
 * Builds a mc_data_entry structure / packet from the given parts. All parameters
 * except the tags are mandatory, and the tags and n_tags can be NULL. It
 * allocates the resulting mde structures (the caller must free it).
 */
static inline int
mc_build_data_entry(struct mc_data_entry **mde, char *key, unsigned int key_len, char *data, unsigned int data_len, struct mc_tag *tags, unsigned int n_tags, time_t exptime, unsigned short int *seq)
{
	unsigned mde_size;
	char *d;

	if (key_len == 0 || data_len == 0)
		return MCREASON_PARAM;
	if ((tags == NULL && n_tags != 0) || (tags != NULL && n_tags == 0))
		return MCREASON_PARAM;
	mde_size = sizeof(struct mc_data_entry) + key_len + data_len + n_tags * sizeof(struct mc_tag);
	*mde = (struct mc_data_entry*)malloc(mde_size);
	if (*mde == NULL)
		return MCREASON_ENOMEM;
	mc_init_packet(*mde, mde_size, (*seq)++, MCMD_PUT_DATA);
	(*mde)->n_tags = n_tags;
	(*mde)->name_size = key_len;
	(*mde)->data_size = data_len;
	(*mde)->exptime = exptime;

	d = (*mde)->data;
	if (tags != NULL) {
		memcpy(d, tags, n_tags * sizeof(struct mc_tag));
		d += n_tags * sizeof(struct mc_tag);
	}
	memcpy(d, key, key_len);
	d += key_len;
	memcpy(d, data, data_len);

	return MCREASON_OK;
}


#ifdef __cplusplus
}
#endif

#endif

==> mdcached-1.0.3.tgz: mdcached-1.0.3/mdcached/mc_protocol.h <==
/*-
 * This file is a part of MDCacheD
 * (c) 2007.-2011. Ivan Voras <ivoras@gmail.com>
 * Released under the 2-clause BSDL
 * $Id: mc_protocol.h 437 2012-03-05 22:08:46Z ivoras $
 */

/*
 * MDCacheD protocol-related definitions.
 */

#ifndef MC_PROTOCOL_H
#define MC_PROTOCOL_H

#include <stdint.h>

#ifdef __cplusplus
extern "C" {
#endif

#ifndef TRUE
#define TRUE 1
#define FALSE (!TRUE)
#endif

#define MC_PROTOCOL_VERSION	1
#define MC_PROTOCOL_VERSION_ID	"$Id: mc_protocol.h 437 2012-03-05 22:08:46Z ivoras $"

/*
 * Depends on: nothing.
 * This file is usable as a client-side header if the MDCACHED_SERVER symbol
 * is not defined.
 */


/*
 * This defines a network protocol so anything that goes over the wire must be
 * declared here.
 */

#define MC_DEFAULT_INET_PORT 	1222
#define MC_BASE_UNIX_SOCKET	"/tmp/mdcached"
#define MC_XSTR(s) #s
#define MC_GET_UNIX_SOCKET_PORT(port)	(MC_BASE_UNIX_SOCKET "." MC_XSTR(port))
#define MC_DEFAULT_UNIX_SOCKET MC_GET_UNIX_SOCKET_PORT(MC_DEFAULT_INET_PORT)

/**
 * Cutoff time in seconds. Time specifications less than this number
 * of seconds will be treated as relative to current time.
 */
#define REL_TIME_CUTOFF (60*60*24*30)

#ifdef TAGS_64BIT
#warning Record tags redefined as 64-bit!
typedef int64_t tag_value_t;		/* tag names and tag values are integers */
typedef int64_t tag_key_t;
#else
typedef int32_t tag_key_t;
typedef int32_t tag_value_t;
#endif

#define MCMD_DATA_ENTRY	MCMD_PUT_DATA

enum MCMD {
	/* commands */
	MCMD_INVALID			= 0,
	MCMD_HANDSHAKE			= 1,
	MCMD_PUT_DATA			= 2,
	MCMD_DEL_DATA_NAME		= 3,
	MCMD_DEL_DATA_BY_TAG_VALUES	= 4,
	MCMD_GET_DATA_NAME		= 5,
	MCMD_GET_DATA_BY_TAG_VALUES	= 6,
	MCMD_GET_ALL_KEYS		= 7,
	MCMD_ATOMIC_CMPSET		= 8,
	MCMD_ATOMIC_ADD			= 9,
	MCMD_ATOMIC_FETCHADD		= 10,
	MCMD_ATOMIC_READANDCLEAR	= 11,
	MCMD_MPUT			= 12,
	MCMD_MGET			= 13,
	MCMD_MDEL			= 14,
	MCMD_TSTACK_PUSH		= 15,
	MCMD_TSTACK_POP			= 16,

	/* responses */
	MCRESP_STATUS			= 100,
	MCRESP_HANDSHAKE		= 101,
	MCRESP_DATA			= 102,
	MCRESP_MULTIDATA		= 103,
	MCRESP_MULTISTRING		= 104,
	MCRESP_ATOMIC			= 105,
	MCRESP_TSTACK_PUSH		= 106,
	MCRESP_TSTACK_POP		= 107
};


enum MCSTATUS {
	MCSTATUS_OK,
	MCSTATUS_ERROR
};


enum MCREASON {
	MCREASON_OK = 0,
	MCREASON_INVALID,
	MCREASON_ENDIANESS,
	MCREASON_VERSION,
	MCREASON_INVALIDCMD,
	MCREASON_CACHE,
	MCREASON_NOTFOUND,
	MCREASON_SERVER,
	MCREASON_PARAM,
	MCREASON_ENOMEM
};

/* Header flags and definitions */
#define MC_ENDIAN_SIGNATURE 	0x1234abcd
#define MC_VER_MAJOR 		0
#define MC_VER_MINOR 		1
#define MC_FLAG_REPLY		0x01

/* Command flags */
#define MCMD_FLAG_ATOMICALL	0x1

/*
 * Command and response packets use the same over the wire format, starting with
 * mc_header and then continuing with the data.
 * Data layout in packets:
 *  - header (mc_header)
 *  - packet specific struct (mc_handshake | mc_add_data | ...)
 *  - variable-length data specific to command
 *    General layout of the variable section is: (not applicable fields are skipped)
 *        - tags (num_tags, tags)
 *        - name (name_length, name)
 *        - data (data_length, data)
 *  - name and data are variable-length, non-0-terminated binary strings.
 *  - name is limited to 64 kB, data to 4 GB.
 */

#ifndef __packed
 #define __packed        __attribute__((__packed__))
#endif


/* Record tag */
struct __packed mc_tag {
	tag_key_t key;
	tag_value_t value;
};


/**
 * Key name simple definition
 */
struct __packed mc_key_name {
	uint16_t name_size;
	char name[];
};


/**
 * Key data simple definition
 */
struct __packed mc_key_data {
	uint32_t data_size;
	char data[];
};


/**
 * Command packet header. This tructure is at the beginning of *every* command
 * packet and response in the client-server protocol.
 */
struct __packed mc_header {
	uint8_t mcmd;		/* Command type */
	uint8_t flags;		/* Command flags */
	uint16_t seq;		/* Command id (sequential) */
	uint32_t total_size;	/* Total size of the command packet, including this header */
};


/** "Handshake" command packet */
struct __packed mc_handshake {
	struct mc_header header;
	int32_t endian_signature;
	uint16_t proto_ver_major;
	uint16_t proto_ver_minor;
};


/** A "simple request" packet. Doesn't carry any data. */
struct __packed mc_simple {
	struct mc_header header;
	uint32_t flags;
	uint32_t param;
};


/**
 * "Add data" command packet. The payload (data field) contains an array of
 * struct mc_tag (n_tags items), followed by the name (of the length name_size),
 * followed by the record value/data (of the length data_size).
 */
struct __packed mc_data_entry {
	struct mc_header header;
	uint16_t n_tags;
	uint16_t name_size;
	uint32_t data_size;
	uint32_t exptime;
	char data[];			/* Dummy placeholder */
};

#define mc_add_data mc_data_entry
#define mc_resp_data mc_data_entry

static inline struct mc_tag* mc_data_entry_tags(struct mc_data_entry *md) {
	return (struct mc_tag*)(md->data);
}

static inline unsigned mc_data_entry_n_tags(struct mc_data_entry *md) {
	return md->n_tags;
}

static inline char* mc_data_entry_key(struct mc_data_entry *md) {
	return md->data + md->n_tags * sizeof(struct mc_tag);
}

static inline unsigned mc_data_entry_key_size(struct mc_data_entry *md) {
	return md->name_size;
}

static inline char* mc_data_entry_value(struct mc_data_entry *md) {
	return md->data + md->n_tags * sizeof(struct mc_tag) + md->name_size;
}

static inline unsigned mc_data_entry_value_size(struct mc_data_entry *md) {
	return md->data_size;
}

/** "Delete data by name" command packet */
struct __packed mc_del_data_name {
	struct mc_header header;
	uint16_t name_size;
	uint16_t _unused;
	char data[];
};


/** "Delete data by tag values" command packet */
struct __packed mc_del_data_by_tag_values {
	struct mc_header header;
	uint32_t n_values;
	tag_key_t tag_key;
	char data[];
};


/** "Get data by name" command packet */
struct __packed mc_get_data_name {
	struct mc_header header;
	uint16_t name_size;
	uint16_t _unused;
	char data[];
};


/** "Get data by tags" command packet */
struct __packed mc_get_data_by_tag_values {
	struct mc_header header;
	uint32_t n_values;
	tag_key_t tag_key;
	char data[];
};


/** A common command packet for simple atomic ops */
struct __packed mc_atomic_op {
	struct mc_header header;
	int64_t arg1;
	int64_t arg2;
	uint16_t name_size;
	uint16_t _unused;
	char name[];
};


/**
 * A multi-put operation: inserts multiple records into the datastore. The data
 * contains n_records repetitions of mc_data_entry records.
 */
struct __packed mc_mput {
	struct mc_header header;
	uint16_t n_records;
	uint16_t flags;
	char data[];
};


/**
 * A multi-get operation: gets multiple records by their names. The data
 * contains n_names repetitions of (uint16_t name length, name bytes)
 */
struct __packed mc_mget {
	struct mc_header header;
	uint16_t n_names;
	uint16_t flags;
	char data[];
};


/**
 * A multi-delete operation: deletes multiple records by their names. The data
 * contains n_names repetitions of (uint16_t name length, name bytes)
 */
struct __packed mc_mdel {
	struct mc_header header;
	uint16_t n_names;
	uint16_t flags;
	char data[];
};


/**
 * The tag-stack push operation. Data contains n_tags repetitions of the struct
 * mc_tag followed by the data of data_size bytes. Response is mc_resp_tstack_push.
 */
struct __packed mc_tstack_push {
	struct mc_header header;
	uint16_t flags;
	uint16_t n_tags;
	uint32_t data_size;
	char data[];
};


/**
 * The tag-stack pop operation. The tag (key and value) references the virtual
 * stack from which the record is popped (returned & removed). Response is mc_resp_
 */
struct __packed mc_tstack_pop {
	struct mc_header header;
	struct mc_tag tag;
	uint16_t flags;
};


/**
 * Status response packet
 */
struct __packed mc_resp_status {
	struct mc_header header;
	uint16_t status;
	uint16_t data;
};


/**
 * "Multi data" response - when a request matches multiple data items. The data
 * payload contains concatenated struct mc_data_entry blobs, unaligned.
 */
struct __packed mc_resp_multidata {
	struct mc_header header;
	uint32_t n_records;
	char data[];
};


/**
 * "Multi string" response - when a command returns an arbitrary list of strings.
 * The data payload consts of a concatenated list of (uint16_t name_len, char name[name_len])
 * records, unaligned. The length is declared as short unsigned integer as this
 * response type is intended to return record keys (key names).
 */
struct __packed mc_resp_multistring {
	struct mc_header header;
	uint32_t n_records;
	uint32_t data_size;
	char data[];
};


/**
 * A simple reply to an atomic op.
 */
struct __packed mc_resp_atomic {
	struct mc_header header;
	int64_t result;
};


/**
 * The response to struct mc_tstack_push. It contains the unique name generated
 * for the record.
 */
struct __packed mc_resp_tstack_push {
	struct mc_header header;
	uint16_t name_size;
	char name[];
};


/**
 * Update the common fields in the packet.
 */
static inline void
mc_init_packet(void *packet, const unsigned int size, const unsigned short int seq, const short int cmd)
{
	struct mc_header *pkt = (struct mc_header*) packet;
	pkt->mcmd = cmd;
	pkt->flags = 0;
	pkt->seq = seq;
	pkt->total_size = size;
}


/**
 * Builds a mc_data_entry structure / packet from the given parts. All parameters
 * except the tags are mandatory, and the tags and n_tags can be NULL. It
 * allocates the resulting mde structures (the caller must free it).
 */
static inline int
mc_build_data_entry(struct mc_data_entry **mde, const char *key, unsigned int key_len, const char *data, unsigned int data_len, const struct mc_tag *tags, unsigned int n_tags, time_t exptime, unsigned short int *seq)
{
	unsigned mde_size;
	char *d;

	if (key_len == 0 || data_len == 0)
		return MCREASON_PARAM;
	if ((tags == NULL && n_tags != 0) || (tags != NULL && n_tags == 0))
		return MCREASON_PARAM;
	mde_size = sizeof(struct mc_data_entry) + key_len + data_len + n_tags * sizeof(struct mc_tag);
	*mde = (struct mc_data_entry*)malloc(mde_size);
	if (*mde == NULL)
		return MCREASON_ENOMEM;
	if (seq != NULL)
		mc_init_packet(*mde, mde_size, (*seq)++, MCMD_PUT_DATA);
	else
		mc_init_packet(*mde, mde_size, 0, MCMD_PUT_DATA);
	(*mde)->n_tags = n_tags;
	(*mde)->name_size = key_len;
	(*mde)->data_size = data_len;
	(*mde)->exptime = exptime;

	d = (*mde)->data;
	if (tags != NULL) {
		memcpy(d, tags, n_tags * sizeof(struct mc_tag));
		d += n_tags * sizeof(struct mc_tag);
	}
	memcpy(d, key, key_len);
	d += key_len;
	memcpy(d, data, data_len);

	return MCREASON_OK;
}


#ifdef __cplusplus
}
#endif

#endif
